@@ -1,15 +1,18 @@
+//! The core information flow analysis.
+
 use std::cell::RefCell;
 
-pub use analysis::{FlowAnalysis, FlowDomain};
-pub use dependencies::{compute_dependencies, compute_dependency_spans, Direction};
 use log::debug;
 use rustc_borrowck::consumers::BodyWithBorrowckFacts;
 use rustc_hir::BodyId;
 use rustc_middle::ty::TyCtxt;
 
+pub use self::{
+  analysis::{FlowAnalysis, FlowDomain},
+  dependencies::{compute_dependencies, compute_dependency_spans, Direction},
+};
 use crate::{
   block_timer,
-  indexed::impls::LocationDomain,
   mir::{
     aliases::Aliases, control_dependencies::ControlDependencies, engine, utils::BodyExt,
   },
@@ -27,6 +30,13 @@ thread_local! {
     RefCell::new(Vec::new());
 }
 
+/// Computes information flow for a MIR body.
+///
+/// The generated data structure essentially is a map from `(Place, Location)` to
+/// `Set<Location>`. For a given place `p` at a location `L`, it says that `p`
+/// is influenced by each location in the set. In static analysis terminology,
+/// this is a flow-sensitive analysis whose domain is the mapping from places to
+/// sets of locations.
 pub fn compute_flow<'a, 'tcx>(
   tcx: TyCtxt<'tcx>,
   body_id: BodyId,
@@ -43,24 +53,16 @@ pub fn compute_flow<'a, 'tcx>(
 
     let def_id = tcx.hir().body_owner_def_id(body_id).to_def_id();
     let aliases = Aliases::build(tcx, def_id, body_with_facts);
+    let location_domain = aliases.location_domain().clone();
 
     let body = &body_with_facts.body;
     let control_dependencies = ControlDependencies::build(body.clone());
     debug!("Control dependencies: {control_dependencies:?}");
 
-    let location_domain = LocationDomain::new(body, &aliases.place_domain);
-
     let results = {
       block_timer!("Flow");
 
-      let analysis = FlowAnalysis::new(
-        tcx,
-        def_id,
-        body,
-        aliases,
-        control_dependencies,
-        location_domain.clone(),
-      );
+      let analysis = FlowAnalysis::new(tcx, def_id, body, aliases, control_dependencies);
       engine::iterate_to_fixpoint(tcx, body, location_domain, analysis)
       // analysis.into_engine(tcx, body).iterate_to_fixpoint()
     };
