@@ -15,20 +15,22 @@ use log::{trace, warn};
 use rustc_data_structures::fx::{FxHashMap as HashMap, FxHashSet as HashSet};
 use rustc_graphviz as dot;
 use rustc_hir::def_id::DefId;
+use rustc_infer::infer::TyCtxtInferExt;
 use rustc_middle::{
   mir::{
     pretty::write_mir_fn,
     visit::{PlaceContext, Visitor},
     MirPass, *,
   },
+  traits::ObligationCause,
   ty::{
-    self, AdtKind, RegionKind, RegionVid, Ty, TyCtxt, TyKind, TyS, TypeFoldable,
-    TypeVisitor,
+    self, AdtKind, RegionKind, RegionVid, Ty, TyCtxt, TyKind, TypeFoldable, TypeVisitor,
   },
 };
 use rustc_mir_dataflow::{fmt::DebugWithContext, graphviz, Analysis, Results};
-use rustc_span::{Pos, Span, Symbol, SyntaxContext};
+use rustc_span::{Pos, Span, SpanData, Symbol, SyntaxContext};
 use rustc_target::abi::VariantIdx;
+use rustc_trait_selection::infer::InferCtxtExt;
 use smallvec::SmallVec;
 
 use crate::extensions::{is_extension_active, MutabilityMode};
@@ -58,7 +60,7 @@ pub fn arg_mut_ptrs<'tcx>(
     .iter()
     .flat_map(|(i, place)| {
       place
-        .interior_pointers(tcx, body, def_id)
+        .interior_pointers(tcx, body, def_id, false)
         .into_iter()
         .flat_map(|(_, places)| {
           places
@@ -178,21 +180,23 @@ impl Visitor<'tcx> for PlaceCollector<'tcx> {
 
   fn visit_rvalue(&mut self, rvalue: &Rvalue<'tcx>, location: Location) {
     match rvalue {
-      Rvalue::Aggregate(box AggregateKind::Adt(def_id, idx, _substs, _, _), ops) => {
+      Rvalue::Aggregate(box AggregateKind::Adt(def_id, idx, substs, _, _), ops) => {
         // In the case of _1 = aggregate { field1: op1, field2: op2, ... }
         // we want to remember which places correspond to which fields so the infoflow
         // analysis can be field-sensitive for constructors.
         let adt_def = self.tcx.adt_def(*def_id);
         let variant = &adt_def.variants[*idx];
-        let places =
-          (0 .. variant.fields.len())
-            .zip(ops.iter())
-            .filter_map(|(field, op)| {
-              let place = op.to_place()?;
-              let field =
-                ProjectionElem::Field(Field::from_usize(field), self.tcx.mk_unit());
-              Some((place, Some(field)))
-            });
+        let places = variant
+          .fields
+          .iter()
+          .enumerate()
+          .zip(ops.iter())
+          .filter_map(|((i, field), op)| {
+            let place = op.to_place()?;
+            let field =
+              ProjectionElem::Field(Field::from_usize(i), field.ty(self.tcx, substs));
+            Some((place, Some(field)))
+          });
         self.places.extend(places);
       }
       _ => self.super_rvalue(rvalue, location),
@@ -285,15 +289,16 @@ pub trait PlaceExt<'tcx> {
     tcx: TyCtxt<'tcx>,
     body: &Body<'tcx>,
     def_id: DefId,
+    shallow: bool,
   ) -> HashMap<RegionVid, Vec<(Place<'tcx>, Mutability)>>;
   fn interior_places(
     &self,
     tcx: TyCtxt<'tcx>,
     body: &Body<'tcx>,
     def_id: DefId,
-    depth_limit: Option<usize>,
   ) -> Vec<Place<'tcx>>;
   fn to_string(&self, tcx: TyCtxt<'tcx>, body: &Body<'tcx>) -> Option<String>;
+  fn normalize(&self, tcx: TyCtxt<'tcx>, def_id: DefId) -> Place<'tcx>;
 }
 
 impl PlaceExt<'tcx> for Place<'tcx> {
@@ -359,6 +364,7 @@ impl PlaceExt<'tcx> for Place<'tcx> {
     tcx: TyCtxt<'tcx>,
     body: &Body<'tcx>,
     def_id: DefId,
+    shallow: bool,
   ) -> HashMap<RegionVid, Vec<(Place<'tcx>, Mutability)>> {
     let ty = self.ty(body.local_decls(), tcx).ty;
     let mut region_collector = CollectRegions {
@@ -370,7 +376,11 @@ impl PlaceExt<'tcx> for Place<'tcx> {
       regions: HashMap::default(),
       places: None,
       types: None,
-      depth_limit: None,
+      stop_at: if shallow {
+        StoppingCondition::AfterRefs
+      } else {
+        StoppingCondition::None
+      },
     };
     region_collector.visit_ty(ty);
     region_collector.regions
@@ -381,7 +391,6 @@ impl PlaceExt<'tcx> for Place<'tcx> {
     tcx: TyCtxt<'tcx>,
     body: &Body<'tcx>,
     def_id: DefId,
-    depth_limit: Option<usize>,
   ) -> Vec<Place<'tcx>> {
     let ty = self.ty(body.local_decls(), tcx).ty;
     let mut region_collector = CollectRegions {
@@ -393,7 +402,7 @@ impl PlaceExt<'tcx> for Place<'tcx> {
       regions: HashMap::default(),
       places: Some(HashSet::default()),
       types: None,
-      depth_limit,
+      stop_at: StoppingCondition::BeforeRefs,
     };
     region_collector.visit_ty(ty);
     region_collector.places.unwrap().into_iter().collect()
@@ -449,6 +458,49 @@ impl PlaceExt<'tcx> for Place<'tcx> {
         .fold(local_name, projection_to_string),
     )
   }
+
+  fn normalize(&self, tcx: TyCtxt<'tcx>, def_id: DefId) -> Place<'tcx> {
+    // Consider a place _1: &'1 <T as SomeTrait>::Foo[2]
+    //   we might encounter this type with a different region, e.g. &'2
+    //   we might encounter this type with a more specific type for the associated type, e.g. &'1 [i32][0]
+    // to account for this variation, we normalize associated types,
+    //   erase regions, and normalize projections
+    let param_env = tcx.param_env(def_id);
+    let place = tcx.erase_regions(*self);
+    let place = tcx.infer_ctxt().enter(|infcx| {
+      infcx
+        .partially_normalize_associated_types_in(
+          ObligationCause::dummy(),
+          param_env,
+          place,
+        )
+        .value
+    });
+
+    let projection = place
+      .projection
+      .into_iter()
+      .filter_map(|elem| match elem {
+        // Map all indexes [i] to [0] since they should be considered equal
+        ProjectionElem::Index(_) | ProjectionElem::ConstantIndex { .. } => {
+          Some(ProjectionElem::Index(Local::from_usize(0)))
+        }
+        // Ignore subslices, they should be treated the same as the
+        // full slice
+        ProjectionElem::Subslice { .. } => None,
+        _ => Some(elem),
+      })
+      .collect::<Vec<_>>();
+
+    Place::make(place.local, &projection, tcx)
+  }
+}
+
+#[derive(Copy, Clone)]
+enum StoppingCondition {
+  None,
+  BeforeRefs,
+  AfterRefs,
 }
 
 struct CollectRegions<'tcx> {
@@ -460,25 +512,15 @@ struct CollectRegions<'tcx> {
   places: Option<HashSet<Place<'tcx>>>,
   types: Option<HashSet<Ty<'tcx>>>,
   regions: HashMap<RegionVid, Vec<(Place<'tcx>, Mutability)>>,
-  depth_limit: Option<usize>,
+  stop_at: StoppingCondition,
 }
 
 impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
   fn visit_ty(&mut self, ty: Ty<'tcx>) -> ControlFlow<Self::BreakTy> {
-    if self
-      .ty_stack
-      .iter()
-      .any(|visited_ty| TyS::same_type(ty, visited_ty))
-    {
+    if self.ty_stack.iter().any(|visited_ty| ty == *visited_ty) {
       return ControlFlow::Continue(());
     }
 
-    if let Some(limit) = self.depth_limit {
-      if self.place_stack.len() > limit {
-        return ControlFlow::Continue(());
-      }
-    }
-
     trace!(
       "exploring {:?} with {ty:?}",
       Place::make(self.local, &self.place_stack, self.tcx)
@@ -546,16 +588,22 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
         self
           .place_stack
           .push(ProjectionElem::Index(Local::from_usize(0)));
-        self.visit_ty(elem_ty);
+        self.visit_ty(*elem_ty);
         self.place_stack.pop();
       }
 
-      TyKind::Ref(region, elem_ty, _) => {
-        self.visit_region(region);
-        self.place_stack.push(ProjectionElem::Deref);
-        self.visit_ty(elem_ty);
-        self.place_stack.pop();
-      }
+      TyKind::Ref(region, elem_ty, _) => match self.stop_at {
+        StoppingCondition::None => {
+          self.visit_region(*region);
+          self.place_stack.push(ProjectionElem::Deref);
+          self.visit_ty(*elem_ty);
+          self.place_stack.pop();
+        }
+        StoppingCondition::AfterRefs => {
+          self.visit_region(*region);
+        }
+        StoppingCondition::BeforeRefs => {}
+      },
 
       TyKind::Closure(_, substs) => {
         self.visit_ty(substs.as_closure().tupled_upvars_ty());
@@ -592,35 +640,38 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
 
   fn visit_region(&mut self, region: ty::Region<'tcx>) -> ControlFlow<Self::BreakTy> {
     trace!("visiting region {region:?}");
-    match region {
-      RegionKind::ReVar(region) => {
-        let mutability = if self
-          .ty_stack
-          .iter()
-          .any(|ty| ty.is_ref() && ty.ref_mutability().unwrap() == Mutability::Not)
-        {
-          Mutability::Not
-        } else {
-          Mutability::Mut
-        };
-
-        let place = Place::make(self.local, &self.place_stack, self.tcx);
-
-        self
-          .regions
-          .entry(*region)
-          .or_default()
-          .push((place, mutability));
-
-        // for initialization setup of Aliases::build
-        if let Some(places) = self.places.as_mut() {
-          places.insert(self.tcx.mk_place_deref(place));
-        }
+    let region = match region.kind() {
+      RegionKind::ReVar(region) => region,
+      RegionKind::ReStatic => RegionVid::from_usize(0),
+      RegionKind::ReErased => {
+        return ControlFlow::Continue(());
       }
-      RegionKind::ReStatic | RegionKind::ReErased => {}
       _ => unreachable!("{:?}: {:?}", self.ty_stack.first().unwrap(), region),
     };
 
+    let mutability = if self
+      .ty_stack
+      .iter()
+      .any(|ty| ty.is_ref() && ty.ref_mutability().unwrap() == Mutability::Not)
+    {
+      Mutability::Not
+    } else {
+      Mutability::Mut
+    };
+
+    let place = Place::make(self.local, &self.place_stack, self.tcx);
+
+    self
+      .regions
+      .entry(region)
+      .or_default()
+      .push((place, mutability));
+
+    // for initialization setup of Aliases::build
+    if let Some(places) = self.places.as_mut() {
+      places.insert(self.tcx.mk_place_deref(place));
+    }
+
     ControlFlow::Continue(())
   }
 }
@@ -733,10 +784,10 @@ impl SpanExt for Span {
   ///  child_spans:    ---      --  -
   ///  output:        -   ------  --
   fn subtract(&self, mut child_spans: Vec<Span>) -> Vec<Span> {
+    child_spans.retain(|s| s.overlaps_inclusive(*self));
+
     let mut outer_spans = vec![];
     if !child_spans.is_empty() {
-      child_spans.retain(|s| s.overlaps_inclusive(*self));
-
       // Output will be sorted
       child_spans = Span::merge_overlaps(child_spans);
 
@@ -827,6 +878,16 @@ impl SpanExt for Span {
   }
 }
 
+pub trait SpanDataExt {
+  fn size(&self) -> u32;
+}
+
+impl SpanDataExt for SpanData {
+  fn size(&self) -> u32 {
+    self.hi.0 - self.lo.0
+  }
+}
+
 // #[derive(Debug)]
 // struct Loop {
 //   header: Location,
